import einops
import torch


def view_as_complex(tensor: torch.Tensor) -> torch.Tensor:
    """Workaround for an einops + torch.view_as_complex() issue

    c.f. https://github.com/arogozhnikov/einops/issues/370

    # works
    a = torch.rand(336, 1, 2)
    print(a.shape, ": ", a.stride(0), a.stride(1), a.stride(2))  # torch.Size([336, 1, 2]) :  2 2 1
    b = torch.view_as_complex(a)

    # errors
    a = torch.rand(size=(336, 2))
    a = einops.rearrange(a, 'b (complex c) -> b c complex', complex=2)
    print(a.shape, ": ", a.stride(0), a.stride(1), a.stride(2))  # torch.Size([336, 1, 2]) :  2 1 1
    a = a.contiguous()
    print(a.shape, ": ", a.stride(0), a.stride(1), a.stride(2))  # torch.Size([336, 1, 2]) :  2 1 1
    b = torch.view_as_complex(a)
    """
    if tensor.shape[-2] == 1:
        tensor = einops.rearrange(tensor, '... 1 complex -> ... complex')
        tensor = torch.view_as_complex(tensor)
        tensor = einops.rearrange(tensor, '... -> ... 1')
    else:
        tensor = torch.view_as_complex(tensor)
    return tensor





import numpy as np
import torch
import einops
import pytest

from torch_image_interpolation import sample_image_3d, insert_into_image_3d


def test_sample_image_3d():
    # basic sanity check only
    image = torch.rand((28, 28, 28))

    # make an arbitrary stack (..., 3) of 3d coords
    arbitrary_shape = (6, 7, 8)
    coords = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    samples = sample_image_3d(image=image, coordinates=coords)
    assert samples.shape == arbitrary_shape


def test_sample_image_3d_complex_input():
    # basic sanity check only
    image = torch.complex(real=torch.rand((28, 28, 28)), imag=torch.rand(28, 28, 28))

    # make an arbitrary stack (..., 3) of 3d coords
    arbitrary_shape = (6, 7, 8)
    coords = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    samples = sample_image_3d(image=image, coordinates=coords)
    assert samples.shape == arbitrary_shape


def test_sample_image_3d_multichannel_input():
    n_channels = 3

    # basic sanity check only
    image = torch.rand((n_channels, 28, 28, 28))

    # make an arbitrary stack (..., 3) of 3d coords
    arbitrary_shape = (6, 7, 8)
    coords = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    samples = sample_image_3d(image=image, coordinates=coords)
    assert samples.shape == (*arbitrary_shape, n_channels)


def test_sample_image_3d_multichannel_complex_input():
    n_channels = 3

    # basic sanity check only
    image = torch.complex(real=torch.rand((28, 28, 28)), imag=torch.rand(28, 28, 28))
    image = einops.repeat(image, 'd h w -> c d h w', c=n_channels)

    # make an arbitrary stack (..., 3) of 3d coords
    arbitrary_shape = (6, 7, 8)
    coords = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    samples = sample_image_3d(image=image, coordinates=coords)
    assert samples.shape == (*arbitrary_shape, n_channels)


def test_insert_into_image_3d():
    image = torch.zeros((28, 28, 28)).float()

    # single value
    value = torch.tensor([5]).float()
    coordinate = torch.tensor([10.5, 14.5, 18.5]).view((1, 3))

    # sample
    image, weights = insert_into_image_3d(value, coordinates=coordinate, image=image, interpolation="trilinear")

    # check value (5) is evenly split over 4 nearest pixels
    expected = einops.repeat(torch.tensor([5 / 8]), '1 -> 2 2 2')
    assert torch.allclose(image[10:12, 14:16, 18:20], expected)

    # check for zeros elsewhere
    assert torch.allclose(image[:10, :14, :18], torch.zeros_like(image[:10, :14, :18]))


def test_insert_into_image_3d_multiple():
    image = torch.zeros((28, 28, 28)).float()

    # multiple values
    values = torch.ones(size=(6, 7, 8)).float()
    coordinates = torch.tensor(np.random.randint(low=0, high=27, size=(6, 7, 8, 3)))

    # sample
    image, weights = insert_into_image_3d(values, coordinates=coordinates, image=image, interpolation="trilinear")

    # check for nonzero value at one point
    sample_point = coordinates[0, 0, 0]
    z, y, x = sample_point
    assert image[z, y, x] > 0


def test_insert_into_image_nearest_interp_3d():
    image = torch.zeros((28, 28, 28)).float()

    # single value
    value = torch.tensor([5]).float()
    coordinate = torch.tensor([10.7, 14.3, 18.9]).view((1, 3))

    # sample
    image, weights = insert_into_image_3d(value, coordinates=coordinate, image=image, interpolation='nearest')

    # check value (5) is added at nearest pixel
    expected = torch.zeros((28, 28, 28)).float()
    expected[11, 14, 19] = 5
    assert torch.allclose(image, expected)


def test_insert_multiple_values_into_multichannel_image_3d_trilinear():
    n_channels = 3
    image = torch.zeros((n_channels, 28, 28, 28)).float()

    # multiple values
    arbitrary_shape = (6, 7, 8)
    values = torch.ones(size=(*arbitrary_shape, n_channels)).float()
    coordinates = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    image, weights = insert_into_image_3d(values, coordinates=coordinates, image=image, interpolation="trilinear")

    # check for nonzero value at one point
    sample_point = coordinates[0, 0, 0]
    z, y, x = sample_point
    assert image[0, z, y, x] > 0

    # check output shapes
    assert image.shape == (n_channels, 28, 28, 28)
    assert weights.shape == (28, 28, 28)


def test_insert_multiple_values_into_multichannel_image_2d_nearest():
    n_channels = 3
    image = torch.zeros((n_channels, 28, 28, 28)).float()

    # multiple values
    arbitrary_shape = (6, 7, 8)
    values = torch.ones(size=(*arbitrary_shape, n_channels)).float()
    coordinates = torch.tensor(np.random.randint(low=0, high=27, size=(*arbitrary_shape, 3)))

    # sample
    image, weights = insert_into_image_3d(values, coordinates=coordinates, image=image, interpolation="nearest")

    # check for nonzero value at one point
    sample_point = coordinates[0, 0, 0]
    z, y, x = sample_point
    assert image[0, z, y, x] > 0

    # check output shapes
    assert image.shape == (n_channels, 28, 28, 28)
    assert weights.shape == (28, 28, 28)


@pytest.mark.parametrize(
    "dtype",
    [torch.float32, torch.float64, torch.complex64, torch.complex128]
)
def test_insert_into_image_2d_type_consistency(dtype):
    image = torch.rand((4, 4, 4), dtype=dtype)
    coords = torch.tensor(np.random.uniform(low=0, high=3, size=(3, 4, 5, 3)))
    values = torch.rand(size=(3, 4, 5), dtype=dtype)
    weights = torch.zeros_like(image, dtype=torch.float64)

    for mode in ['bilinear', 'nearest']:
        image, weights = insert_into_image_3d(
            values,
            image=image,
            weights=weights,
            coordinates=coords,
            interpolation=mode,
        )
        assert image.dtype == dtype
        assert weights.dtype == torch.float64


def test_insert_into_image_3d_type_error():
    image = torch.rand((4, 4, 4), dtype=torch.complex64)
    coords = torch.tensor(np.random.uniform(low=0, high=3, size=(3, 4, 5, 3)))
    values = torch.rand(size=(3, 4, 5), dtype=torch.complex128)
    with pytest.raises(ValueError):
        insert_into_image_3d(
            values,
            image=image,
            coordinates=coords,
        )



from typing import Literal

import einops
import torch
import torch.nn.functional as F

from torch_image_interpolation import utils
from .grid_sample_utils import array_to_grid_sample


def sample_image_2d(
    image: torch.Tensor,
    coordinates: torch.Tensor,
    interpolation: Literal['nearest', 'bilinear', 'bicubic'] = 'bilinear',
) -> torch.Tensor:
    """Sample a 2D image with a specific interpolation mode.

    Parameters
    ----------
    image: torch.Tensor
        `(h, w)` image or `(c, h, w)` multichannel image.
    coordinates: torch.Tensor
        `(..., 2)` array of coordinates at which `image` should be sampled.
        - Coordinates are ordered `yx` and are positions in the `h` and `w` dimensions respectively.
        - Coordinates span the range `[0, N-1]` for a dimension of length N.
    interpolation: Literal['nearest', 'bilinear', 'bicubic']
        Interpolation mode for image sampling.

    Returns
    -------
    samples: torch.Tensor
        `(..., )` or `(..., c)` array of samples from `image`.
    """
    device = coordinates.device

    if image.ndim not in (2, 3):
        raise ValueError(f'image should have shape (h, w) or (c, h, w), got {image.shape}')

    # keep track of a few image properties
    input_image_is_complex = torch.is_complex(image)
    input_image_is_multichannel = image.ndim == 3

    # coerce single channel to multi-channel
    if input_image_is_multichannel is False:
        image = einops.rearrange(image, 'h w -> 1 h w')

    # set up for sampling with torch.nn.functional.grid_sample
    # shape (..., 2) -> (n, 2)
    coordinates, ps = einops.pack([coordinates], pattern='* yx')
    n_samples = coordinates.shape[0]
    h, w = image.shape[-2:]

    # handle complex input
    if input_image_is_complex:
        # cannot sample complex tensors directly with grid_sample
        # c.f. https://github.com/pytorch/pytorch/issues/67634
        # workaround: treat real and imaginary parts as separate channels
        image = torch.view_as_real(image)
        image = einops.rearrange(image, 'c h w complex -> (complex c) h w')

    # torch.nn.functional.grid_sample is set up for sampling grids
    # here we view our image as a batch of n_samples multi-channel images
    # then sample a batch of (1x1) grids
    # this enables sampling arbitrarily shaped arrays of coords
    image = einops.repeat(image, 'c h w -> b c h w', b=n_samples)
    coordinates = einops.rearrange(coordinates, 'b yx -> b 1 1 yx')  # b h w yx

    # take the samples
    samples = F.grid_sample(
        input=image,
        grid=array_to_grid_sample(coordinates, array_shape=image.shape[-2:]),
        mode=interpolation,
        padding_mode='border',  # this increases sampling fidelity at edges
        align_corners=True,
    )

    # reconstruct complex valued samples if required
    if input_image_is_complex is True:
        samples = einops.rearrange(samples, 'b (complex c) 1 1 -> b c complex', complex=2)
        samples = utils.view_as_complex(samples.contiguous())  # (b, c)
    else:
        samples = einops.rearrange(samples, 'b c 1 1 -> b c')

    # set samples from outside of image to zero explicitly
    coordinates = einops.rearrange(coordinates, 'b 1 1 yx -> b yx')
    image_shape = torch.as_tensor(image.shape[-2:]).to(device)
    inside = torch.logical_and(coordinates >= 0, coordinates <= image_shape - 1)
    inside = torch.all(inside, dim=-1)  # (b,)
    samples[~inside] *= 0

    # pack samples back into the expected shape
    # (b, c) -> (..., c)
    [samples] = einops.unpack(samples, pattern='* c', packed_shapes=ps)

    # ensure output has correct shape
    # - (...) if input image was single channel
    # - (..., c) if input image was multi-channel
    if input_image_is_multichannel is False:
        samples = einops.rearrange(samples, '... 1 -> ...')  # drop channel dim

    return samples


def insert_into_image_2d(
    values: torch.Tensor,
    coordinates: torch.Tensor,
    image: torch.Tensor,
    weights: torch.Tensor | None = None,
    interpolation: Literal['nearest', 'bilinear'] = 'bilinear',
) -> tuple[torch.Tensor, torch.Tensor]:
    """Insert values into a 2D image with bilinear interpolation.

    Parameters
    ----------
    values: torch.Tensor
        `(...)` or `(..., c)` array of values to be inserted into `image`.
    coordinates: torch.Tensor
        `(..., 2)` array of 2D coordinates for each value in `data`.
        - Coordinates are ordered `yx` and are positions in the `h` and `w` dimensions respectively.
        - Coordinates span the range `[0, N-1]` for a dimension of length N.
    image: torch.Tensor
        `(h, w)` or `(c, h, w)` array containing the image into which data will be inserted.
    weights: torch.Tensor | None
        `(h, w)` array containing weights associated with each pixel in `image`.
        This is useful for tracking weights across multiple calls to this function.
    interpolation: Literal['nearest', 'bilinear']
        Interpolation mode used for adding data points to grid.

    Returns
    -------
    image, weights: tuple[torch.Tensor, torch.Tensor]
        The image and weights after updating with data from `data` at `coordinates`.
    """
    # keep track of a few properties of the inputs
    input_image_is_multichannel = image.ndim == 3
    h, w = image.shape[-2:]

    # validate inputs
    values_shape = values.shape[:-1] if input_image_is_multichannel else values.shape
    coordinates_shape, coordinates_ndim = coordinates.shape[:-1], coordinates.shape[-1]

    if values_shape != coordinates_shape:
        raise ValueError('One coordinate pair is required for each value in data.')
    if coordinates_ndim != 2:
        raise ValueError('Coordinates must be 2D with shape (..., 2).')
    if image.dtype != values.dtype:
        raise ValueError('Image and values must have the same dtype.')

    if weights is None:
        weights = torch.zeros(size=(h, w), dtype=torch.float32, device=image.device)

    # add channel dim to both image and values if input image is not multichannel
    if not input_image_is_multichannel:
        image = einops.rearrange(image, 'h w -> 1 h w')
        values = einops.rearrange(values, '... -> ... 1')

    # linearise data and coordinates
    values, _ = einops.pack([values], pattern='* c')
    coordinates, _ = einops.pack([coordinates], pattern='* yx')
    coordinates = coordinates.float()

    # only keep data and coordinates inside the image
    image_shape = torch.tensor((h, w), device=image.device, dtype=torch.float32)
    upper_bound = image_shape - 1
    idx_inside = (coordinates >= 0) & (coordinates <= upper_bound)
    idx_inside = torch.all(idx_inside, dim=-1)
    values, coordinates = values[idx_inside], coordinates[idx_inside]

    # splat data onto grid
    if interpolation == 'nearest':
        image, weights = _insert_nearest_2d(values, coordinates, image, weights)
    if interpolation == 'bilinear':
        image, weights = _insert_linear_2d(values, coordinates, image, weights)

    # ensure correct output image shape
    # single channel input -> (h, w)
    # multichannel input -> (c, h, w)
    if not input_image_is_multichannel:
        image = einops.rearrange(image, '1 h w -> h w')

    return image, weights


def _insert_nearest_2d(
    data,  # (b, c)
    coordinates,  # (b, yx)
    image,  # (c, h, w)
    weights  # (h, w)
):
    # b is number of data points to insert per channel, c is number of channels
    b, c = data.shape

    # flatten data to insert values for all channels with one call to _index_put()
    data = einops.rearrange(data, 'b c -> b c')

    # find nearest voxel for each coordinate
    coordinates = torch.round(coordinates).long()
    idx_h, idx_w = einops.rearrange(coordinates, 'b yx -> yx b')

    # insert ones into weights image (h, w) at each position
    w = torch.ones(size=(b, 1), device=weights.device, dtype=weights.dtype)

    # setup indices for insertion
    idx_c = torch.arange(c, device=coordinates.device, dtype=torch.long)
    idx_c = einops.rearrange(idx_c, 'c -> 1 c')
    idx_h = einops.rearrange(idx_h, 'b -> b 1')
    idx_w = einops.rearrange(idx_w, 'b -> b 1')

    # insert image data and weights
    image.index_put_(indices=(idx_c, idx_h, idx_w), values=data, accumulate=True)
    weights.index_put_(indices=(idx_h, idx_w), values=w, accumulate=True)
    return image, weights


def _insert_linear_2d(
    data,  # (b, c)
    coordinates,  # (b, yx)
    image,  # (c, h, w)
    weights  # (h, w)
):
    # b is number of data points to insert per channel, c is number of channels
    b, c = data.shape

    # cache corner coordinates for each value to be inserted
    #     C10---C11
    #      |  P  |
    #     C00---C01
    coordinates = einops.rearrange(coordinates, 'b yx -> yx b')
    y0, x0 = torch.floor(coordinates)
    y1, x1 = torch.ceil(coordinates)

    # populate arrays of corner indices
    idx_h = torch.empty(size=(b, 2, 2), dtype=torch.long, device=image.device)
    idx_w = torch.empty(size=(b, 2, 2), dtype=torch.long, device=image.device)

    idx_h[:, 0, 0], idx_w[:, 0, 0] = y0, x0  # C00
    idx_h[:, 0, 1], idx_w[:, 0, 1] = y0, x1  # C01
    idx_h[:, 1, 0], idx_w[:, 1, 0] = y1, x0  # C10
    idx_h[:, 1, 1], idx_w[:, 1, 1] = y1, x1  # C11

    # calculate linear interpolation weights for each corner
    y, x = coordinates
    ty, tx = y - y0, x - x0  # fractional position between corners
    w = torch.empty(size=(b, 2, 2), device=image.device, dtype=weights.dtype)
    w[:, 0, 0] = (1 - ty) * (1 - tx)   # C00
    w[:, 0, 1] = (1 - ty) * tx         # C01
    w[:, 1, 0] = ty * (1 - tx)         # C10
    w[:, 1, 1] = ty * tx               # C11

    # make sure indices broadcast correctly
    idx_c = torch.arange(c, device=coordinates.device, dtype=torch.long)
    idx_c = einops.rearrange(idx_c, 'c -> 1 c 1 1')
    idx_h = einops.rearrange(idx_h, 'b h w -> b 1 h w')
    idx_w = einops.rearrange(idx_w, 'b h w -> b 1 h w')

    # insert weighted data and weight values at each corner across all channels
    # make sure to do atomic adds
    data = einops.rearrange(data, 'b c -> b c 1 1')
    w = einops.rearrange(w, 'b h w -> b 1 h w')
    image.index_put_(
        indices=(idx_c, idx_h, idx_w),
        values=data * w.to(data.dtype),
        accumulate=True
    )
    weights.index_put_(indices=(idx_h, idx_w), values=w, accumulate=True)

    return image, weights



from .image_interpolation_1d import sample_image_1d, insert_into_image_1d
from .image_interpolation_2d import sample_image_2d, insert_into_image_2d
from .image_interpolation_3d import sample_image_3d, insert_into_image_3d



from typing import Literal

import einops
import torch
import torch.nn.functional as F

from torch_image_interpolation import utils
from torch_image_interpolation.grid_sample_utils import array_to_grid_sample


def sample_image_1d(
        image: torch.Tensor,
        coordinates: torch.Tensor,
        interpolation: Literal['nearest', 'linear', 'cubic'] = 'linear',
) -> torch.Tensor:
    """Sample a 1D image (vector) with a specific interpolation mode.

    Parameters
    ----------
    image: torch.Tensor
        `(w, )` image or `(c, w)` multichannel image.
    coordinates: torch.Tensor
        `(..., )` array of coordinates at which `image` should be sampled.
        - Coordinates are positions in the `w` dimension.
        - Coordinates span the range `[0, N-1]` for a dimension of length N.
    interpolation: Literal['nearest', 'linear', 'cubic']
        Interpolation mode for image sampling.

    Returns
    -------
    samples: torch.Tensor
        `(..., )` or `(..., c)` array of samples from `image`.
    """
    if image.ndim not in (1, 2):
        raise ValueError(f'image should have shape (w, ) or (c, w), got {image.shape}')

    # keep track of a few image properties
    input_image_is_complex = torch.is_complex(image)
    input_image_is_multichannel = image.ndim == 2

    # coerce single channel to multi-channel
    if input_image_is_multichannel is False:
        image = einops.rearrange(image, 'w -> 1 w')  # (c, w)

    # set up for sampling with torch.nn.functional.grid_sample
    # coordinates: (..., ) -> (n, )
    coordinates, ps = einops.pack([coordinates], pattern='*')
    n_samples = coordinates.shape[0]
    w = image.shape[-1]

    # handle complex input
    if input_image_is_complex:
        # cannot sample complex tensors directly with grid_sample
        # c.f. https://github.com/pytorch/pytorch/issues/67634
        # workaround: treat real and imaginary parts as separate channels
        image = torch.view_as_real(image)
        image = einops.rearrange(image, 'c w complex -> (complex c) w')

    # For grid_sample to work with 1D, we need to add a dummy height dimension
    # and treat it as a 2D image with height=2
    image = einops.repeat(image, 'c w -> c h w', h=2)

    # Create batch of images, one per sample
    image = einops.repeat(image, 'c h w -> b c h w', b=n_samples)

    # Create grid coordinates
    # We need to add a dummy y coordinate (set to 0) to make it work with grid_sample
    dummy_y = torch.zeros_like(coordinates)
    coords_2d = einops.rearrange([dummy_y, coordinates], "yx b -> b 1 1 yx")  # (b, 1, 1, 2)

    # Take the samples
    # We need to convert the coordinates to grid_sample format
    coords_2d = array_to_grid_sample(coords_2d, array_shape=(2, w))
    if interpolation == 'linear':
        mode = 'bilinear'
    elif interpolation == 'cubic':
        mode = 'bicubic'
    else:
        mode = interpolation
    samples = F.grid_sample(
        input=image,
        grid=coords_2d,
        mode=mode,
        padding_mode='border',  # this increases sampling fidelity at edges
        align_corners=True,
    )

    # Reconstruct complex valued samples if required
    if input_image_is_complex is True:
        samples = einops.rearrange(samples, 'b (complex c) 1 1 -> b c complex', complex=2)
        samples = utils.view_as_complex(samples.contiguous())  # (b, c)
    else:
        samples = einops.rearrange(samples, 'b c 1 1 -> b c')

    # Set samples from outside of image to zero explicitly
    inside = torch.logical_and(coordinates >= 0, coordinates <= w - 1)
    samples[~inside] *= 0

    # Pack samples back into the expected shape
    # (b, c) -> (..., c)
    [samples] = einops.unpack(samples, pattern='* c', packed_shapes=ps)

    # Ensure output has correct shape
    # - (...) if input image was single channel
    # - (..., c) if input image was multi-channel
    if input_image_is_multichannel is False:
        samples = einops.rearrange(samples, '... 1 -> ...')  # drop channel dim

    return samples


def insert_into_image_1d(
        values: torch.Tensor,
        coordinates: torch.Tensor,
        image: torch.Tensor,
        weights: torch.Tensor | None = None,
        interpolation: Literal['nearest', 'linear'] = 'linear',
) -> tuple[torch.Tensor, torch.Tensor]:
    """Insert values into a 1D image with linear interpolation.

    Parameters
    ----------
    values: torch.Tensor
        `(...)` or `(..., c)` array of values to be inserted into `image`.
    coordinates: torch.Tensor
        `(..., 1)` array of 1D coordinates for each value in `data`.
        - Coordinates span the range `[0, N-1]` for a dimension of length N.
    image: torch.Tensor
        `(w)` or `(c, w)` array containing the image into which data will be inserted.
    weights: torch.Tensor | None
        `(w)` array containing weights associated with each pixel in `image`.
        This is useful for tracking weights across multiple calls to this function.
    interpolation: Literal['nearest', 'linear']
        Interpolation mode used for adding data points to grid.

    Returns
    -------
    image, weights: tuple[torch.Tensor, torch.Tensor]
        The image and weights after updating with data from `data` at `coordinates`.
    """
    # keep track of a few properties of the inputs
    input_image_is_multichannel = image.ndim == 2
    w = image.shape[-1]

    # validate inputs
    values_shape = values.shape[:-1] if input_image_is_multichannel else values.shape

    if values_shape != coordinates.shape:
        raise ValueError('One coordinate is required for each value in data.')
    if image.dtype != values.dtype:
        raise ValueError('Image and values must have the same dtype.')

    if weights is None:
        weights = torch.zeros(size=(w,), dtype=torch.float32, device=image.device)

    # add channel dim to both image and values if input image is not multichannel
    if not input_image_is_multichannel:
        image = einops.rearrange(image, 'w -> 1 w')
        values = einops.rearrange(values, '... -> ... 1')

    # linearise data and coordinates
    values, _ = einops.pack([values], pattern='* c')  # (b, c)
    coordinates, _ = einops.pack([coordinates], pattern='*')  # (b, )
    coordinates = coordinates.float()

    # only keep data and coordinates inside the image
    idx_inside = (coordinates >= 0) & (coordinates <= w - 1)
    values, coordinates = torch.atleast_2d(values[idx_inside]), coordinates[idx_inside]

    # splat data onto grid
    if interpolation == 'nearest':
        image, weights = _insert_nearest_1d(values, coordinates, image, weights)
    if interpolation == 'linear':
        image, weights = _insert_linear_1d(values, coordinates, image, weights)

    # ensure correct output image shape
    # single channel input -> (w)
    # multichannel input -> (c, w)
    if not input_image_is_multichannel:
        image = einops.rearrange(image, '1 w -> w')

    return image, weights


def _insert_nearest_1d(
        data,  # (b, c)
        coordinates,  # (b,)
        image,  # (c, w)
        weights  # (w)
):
    # b is number of data points to insert per channel, c is number of channels
    b, c = data.shape

    # find nearest voxel for each coordinate
    coordinates = torch.round(coordinates).long()

    # insert ones into weights image (w) at each position
    w = torch.ones(size=(b, 1), device=weights.device, dtype=weights.dtype)

    # setup indices for insertion
    idx_c = torch.arange(c, device=coordinates.device, dtype=torch.long)
    idx_c = einops.rearrange(idx_c, 'c -> 1 c')
    idx_x = einops.rearrange(coordinates, 'b -> b 1')

    # insert image data and weights
    image.index_put_(indices=(idx_c, idx_x), values=data, accumulate=True)
    weights.index_put_(indices=(idx_x,), values=w, accumulate=True)
    return image, weights


def _insert_linear_1d(
        data,  # (b, c)
        coordinates,  # (b,)
        image,  # (c, w)
        weights  # (w)
):
    # b is number of data points to insert per channel, c is number of channels
    b, c = data.shape

    # cache corner coordinates for each value to be inserted
    # C0---P---C1
    x0 = torch.floor(coordinates).long()
    x1 = torch.ceil(coordinates).long()

    # calculate linear interpolation weights for each corner
    tx = coordinates - x0.float()  # fractional position between corners
    w = torch.empty(size=(b, 2), device=image.device, dtype=weights.dtype)
    w[:, 0] = 1 - tx  # C0
    w[:, 1] = tx  # C1

    # Set up indices for both corners
    idx_x = torch.stack([x0, x1], dim=1)  # (b, 2)

    # make sure indices broadcast correctly
    idx_c = torch.arange(c, device=coordinates.device, dtype=torch.long)
    idx_c = einops.rearrange(idx_c, 'c -> 1 c 1')
    idx_x = einops.rearrange(idx_x, 'b x -> b 1 x')

    # insert weighted data and weight values at each corner across all channels
    data = einops.rearrange(data, 'b c -> b c 1')
    w = einops.rearrange(w, 'b x -> b 1 x')
    image.index_put_(
        indices=(idx_c, idx_x),
        values=data * w.to(data.dtype),
        accumulate=True
    )
    weights.index_put_(indices=(idx_x,), values=w, accumulate=True)

    return image, weights
